
/**
 * Copyright 2012-2013 Clint Combs
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.newhart

import scala.util.DynamicVariable
import java.util.{Date => JDate}
import com.weiglewilczek.slf4s.Logging
import com.mongodb.casbah.Imports._

trait AuditAPI {
  def withAudit[R](name: String)(block: => R)(implicit delivery: Delivery):R = {
    AuditAPI._withAudit(name) { block }
  }

  def withNewAudit[R](name: String)(block: => R)(implicit delivery: Delivery):R = {
    AuditAPI._withNewAudit(name) { block }
  }

  def auditOrigin(s: String) = AuditAPI._defaultAuditOrigin(s)
  def auditOriginKey(s: String) = AuditAPI._defaultAuditOriginKey(s)
  def auditOriginKeyType(s: String) = AuditAPI._defaultAuditOriginKeyType(s)
  def auditOriginKeyType[T](c: Class[T]) = AuditAPI._defaultAuditOriginKeyType(c.getName)
  def auditOriginHostname(s: String) = AuditAPI._defaultAuditOriginHostname(s)
  def auditCriticality(c: Criticality.Value) = AuditAPI._defaultAuditCriticality(c)
  def auditTS(ts: JDate) = AuditAPI._defaultAuditTS(ts)
  def auditCreateTS(ts: JDate) = AuditAPI._defaultAuditCreateTS(ts)
  def auditUpdateTS(ts: JDate) = AuditAPI._defaultAuditUpdateTS(ts)
  def auditMsg(s: String) = AuditAPI._defaultAuditMsg(s)
  def auditError(e: Error) = AuditAPI._defaultAuditError(e)
  def auditWarning(w: Warning) = AuditAPI._defaultAuditWarning(w)
  def auditLabel(label: String) = AuditAPI._defaultAuditLabel(label)
  def auditData(d: Tuple2[String, Any]) = AuditAPI._defaultAuditData(d)
  def auditData(d: Map[String, Any]) = AuditAPI._defaultAuditData(d)

  def auditOrigin(name: String, s: String) = AuditAPI._auditOrigin(name, s)
  def auditOriginKey(name: String, s: String) = AuditAPI._auditOriginKey(name, s)
  def auditOriginKeyType(name: String, s: String) = AuditAPI._auditOriginKeyType(name, s)
  def auditOriginKeyType[T](name: String, c: Class[T]) = AuditAPI._auditOriginKeyType(name, c.getName)
  def auditOriginHostname(name: String, s: String) = AuditAPI._auditOriginHostname(name, s)
  def auditCriticality(name: String, c: Criticality.Value) = AuditAPI._auditCriticality(name, c)
  def auditTS(name: String, ts: JDate) = AuditAPI._auditTS(name, ts)
  def auditCreateTS(name: String, ts: JDate) = AuditAPI._auditCreateTS(name, ts)
  def auditUpdateTS(name: String, ts: JDate) = AuditAPI._auditUpdateTS(name, ts)
  def auditMsg(name: String, s: String) = AuditAPI._auditMsg(name, s)
  def auditError(name: String, e: Error) = AuditAPI._auditError(name, e)
  def auditWarning(name: String, w: Warning) = AuditAPI._auditWarning(name, w)
  def auditLabel(name: String, label: String) = AuditAPI._auditLabel(name, label)
  def auditData(name: String, d: Tuple2[String, Any]) = AuditAPI._auditData(name, d)
  def auditData(name: String, d: Map[String, Any]) = AuditAPI._auditData(name, d)
}

/**
 * An Entry builders frame containing an optional default audit name and a Map of audit names to
 * Entry.Builder instances.
 */
// TODO: consider renaming the map field to builderMap
case class EntryBuildersFrame(defaultAuditName: Option[String], map: Map[String, Entry.Builder])

object AuditAPI extends Logging {

  private lazy val entryBuildersFrame: DynamicVariable[EntryBuildersFrame] =
    new DynamicVariable(EntryBuildersFrame(None, Map()))

  /**
   * Return a new or existing EntryBuildersFrame along with a Boolean representing whether
   * or not the frame is new.  A new frame is only created if the old one does not contain
   * the auditName.
   */
  private def entryBuildersFrame(auditName: String): (EntryBuildersFrame, Boolean) = {
    val frame = entryBuildersFrame.value
    if (frame.map.contains(auditName))
      (frame, false)
    else
      (EntryBuildersFrame(Some(auditName), frame.map + (auditName -> Entry.builder())), true)
  }

  /**
   * Create a new EntryBuildersFrame.
   */
  private def newEntryBuildersFrame(auditName: String): EntryBuildersFrame = {
    val frame = entryBuildersFrame.value
    EntryBuildersFrame(Some(auditName), frame.map + (auditName -> Entry.builder()))
  }

  private def _withAudit[R](auditName: String)(block: => R)(implicit delivery: Delivery):R = {
    val (frame, isNewFrame) = entryBuildersFrame(auditName)
    if (isNewFrame) {
      entryBuildersFrame.withValue(frame) {
        val result = try {
          block
        } catch {
          // TODO: carefully consider if the catch should be included here or it should be delegated to @Audit
          case ex: Exception => frame.map(auditName).errors +=
            Error(Some(ex.toString), None, Some(buildStackTraceString(ex)))
          send(auditName, frame.map(auditName))
          throw ex
        }
        send(auditName, frame.map(auditName))
        result
      }
    } else {
      try {
        block
      } catch {
        // TODO: carefully consider if the catch should be included here or it should be delegated to @Audit
        case ex: Exception => frame.map(auditName).errors +=
          Error(Some(ex.toString), None, Some(buildStackTraceString(ex)))
        throw ex
      }
      // this is not a new frame, do not write an audit Entry here
    }
  }

  private def _withNewAudit[R](auditName: String)(block: => R)(implicit delivery: Delivery):R = {
    val frame = newEntryBuildersFrame(auditName)
    entryBuildersFrame.withValue(frame) {
      val result = try {
        block
      } catch {
        // TODO: carefully consider if the catch should be included here or it should be delegated to @Audit
        case ex: Exception => entryBuildersFrame.value.map(auditName).errors +=
          Error(Some(ex.toString), None, Some(buildStackTraceString(ex)))
        send(auditName, frame.map(auditName))
        throw ex
      }
      send(auditName, frame.map(auditName))
      result
    }
  }

  def send(auditName: String, eb: Entry.Builder)(implicit delivery: Delivery) = {
    try {
      val e = completeEntry(eb)
      delivery.send(auditName, e)
    } catch {
      case ex: Exception => logger.error("error writing entry from builder: " + eb, ex)
    }
  }

  /** Complete an Entry before it is written to an audit trail. */
  def completeEntry(eb: Entry.Builder):Entry = {
    eb.auditTS = Entry.now()
    eb.result()
  }

  /** Update the Entry.Builder for the given audit name or the default. */
  private def updateFrameEntryBuilder[R](auditName: Option[String], f: (Entry.Builder) => R) = {

    def defaultFrameAuditName(): Option[String] =
      entryBuildersFrame.value.defaultAuditName

    auditName orElse defaultFrameAuditName map { name =>
      f(entryBuildersFrame.value.map(name))
    }
  }

  private def _defaultAuditError(e: Error) = {
    updateFrameEntryBuilder(None, { _.errors += e })
  }

  private def _defaultAuditWarning(w: Warning) = {
    updateFrameEntryBuilder(None, { _.warnings += w })
  }

  private def _defaultAuditLabel(label: String) = {
    updateFrameEntryBuilder(None, { _.labels += label })
  }

  private def _defaultAuditOrigin(s: String) = {
    updateFrameEntryBuilder(None, { _.origin = s })
  }

  private def _defaultAuditOriginKey(s: String) = {
    updateFrameEntryBuilder(None, { _.originKey = s })
  }

  private def _defaultAuditOriginKeyType(s: String) = {
    updateFrameEntryBuilder(None, { _.originKeyType = s })
  }

  private def _defaultAuditOriginHostname(s: String) = {
    updateFrameEntryBuilder(None, { _.originHostname = s })
  }

  private def _defaultAuditCriticality(c: Criticality.Value) = {
    updateFrameEntryBuilder(None, { _.criticality = c })
  }

  private def _defaultAuditTS(ts: JDate) = {
    updateFrameEntryBuilder(None, { _.auditTS = ts })
  }

  private def _defaultAuditCreateTS(ts: JDate) = {
    updateFrameEntryBuilder(None, { _.createTS = ts })
  }

  private def _defaultAuditUpdateTS(ts: JDate) = {
    updateFrameEntryBuilder(None, { _.updateTS = ts })
  }

  private def _defaultAuditMsg(s: String) = {
    updateFrameEntryBuilder(None, { _.msg = s })
  }

  private def _defaultAuditData(t: Tuple2[String, Any]) = {
    updateFrameEntryBuilder(None, { _.data += (t._1 -> t._2) })
  }

  private def _defaultAuditData(d: Map[String, Any]) = {
    updateFrameEntryBuilder(None, { builder => d map { m => builder.data += m._1 -> m._2 } })
  }

  private def _auditError(name: String, e: Error) = {
    updateFrameEntryBuilder(Some(name), { _.errors += e })
  }

  private def _auditWarning(name: String, w: Warning) = {
    updateFrameEntryBuilder(Some(name), { _.warnings += w })
  }

  private def _auditLabel(name: String, label: String) = {
    updateFrameEntryBuilder(Some(name), { _.labels += label })
  }

  private def _auditOrigin(name: String, s: String) = {
    updateFrameEntryBuilder(Some(name), { _.origin = s })
  }

  private def _auditOriginKey(name: String, s: String) = {
    updateFrameEntryBuilder(Some(name), { _.originKey = s })
  }

  private def _auditOriginKeyType(name: String, s: String) = {
    updateFrameEntryBuilder(Some(name), { _.originKeyType = s })
  }

  private def _auditOriginHostname(name: String, s: String) = {
    updateFrameEntryBuilder(Some(name), { _.originHostname = s })
  }

  private def _auditCriticality(name: String, c: Criticality.Value) = {
    updateFrameEntryBuilder(Some(name), { _.criticality = c })
  }

  private def _auditTS(name: String, ts: JDate) = {
    updateFrameEntryBuilder(Some(name), { _.auditTS = ts })
  }

  private def _auditCreateTS(name: String, ts: JDate) = {
    updateFrameEntryBuilder(Some(name), { _.createTS = ts })
  }

  private def _auditUpdateTS(name: String, ts: JDate) = {
    updateFrameEntryBuilder(Some(name), { _.updateTS = ts })
  }

  private def _auditMsg(name: String, s: String) = {
    updateFrameEntryBuilder(Some(name), { _.msg = s })
  }

  private def _auditData(name: String, t: Tuple2[String, Any]) = {
    updateFrameEntryBuilder(Some(name), { _.data += (t._1 -> t._2) })
  }

  private def _auditData(name: String, d: Map[String, Any]) = {
    updateFrameEntryBuilder(Some(name), { builder => d map { m => builder.data += m._1 -> m._2 } })
  }
}
