
/**
 * Copyright 2012-2013 Clint Combs
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.newhart.akka

import org.newhart._
import _root_.akka.actor.{Actor,ActorRef,ActorSystem,Props}
import org.newhart.{Delivery,Entry,Error,Warning}
import java.util.{Date => JDate}
import java.util.UUID
import com.mongodb.casbah.Imports._
import com.typesafe.config.ConfigFactory

trait AkkaDelivery extends Delivery {

  def send(name: String, e: Entry) = {
    AkkaDelivery.actor ! EntryMessage(name, e)
  }

  @deprecated(message = "Please use send instead of writeEntry.", since = "0.4.5")
  def writeEntry(collectionName: String, e: Entry) = {
    send(name = collectionName, e = e)
  }
}

// TODO: send via Actors
// TODO: setup shutdown mechanism
// TODO: make it possible to use durable mailbox file implementation
// TODO: check getLastError if priority is major or critical(?);
//       maybe this should be configurable via some Criticality policy implementation

private class ConcreteAkkaDelivery() extends AkkaDelivery

object AkkaDelivery {

  // TODO: change this; returning the same Actor every time is not supported
  // http://doc.akka.io/docs/akka/2.0.4/scala/actors.html
  // http://doc.akka.io/docs/akka/2.0.4/general/supervision.html#supervision-restart
  lazy val actor: ActorRef = start()

  private lazy val config = ConfigFactory.load()

  // TODO: this fallback configuration currently fails because "newhart" is not being found in the config file
  //private lazy val system = ActorSystem("newhart-akka-delivery", config.getConfig("newhart").withFallback(config))
  private lazy val system = ActorSystem("newhart-akka-delivery", config)

  private val deliveryActorName = "deliveryActor"
  private val dispatcherName = "newhart-delivery-dispatcher"

  private class AkkaDeliveryShutdown extends Runnable {
    def run() {
      shutdown
    }
  }

  private def start(): ActorRef = {
    implicit val con = MongoConnection("localhost")
    val a = system.actorOf(
      Props(new DeliveryActor()).withDispatcher(dispatcherName), name = deliveryActorName)
    Runtime.getRuntime.addShutdownHook {
      new Thread(new AkkaDeliveryShutdown())
    }
    a
  }

  // TODO: use this for integrating into runtime environments
  // as the callback when it's time to shutdown
  def shutdown() = {
    system.shutdown
  }
}

object AkkaDeliveryExample1 extends App with AuditAPI {

  implicit val delivery = new AkkaDelivery() { }

  for { i <- 1 to 3 } {
    withAudit("users") {
      auditOrigin("AkkaDeliveryExample1")
      auditOriginKey(UUID.randomUUID.toString)
      auditOriginKeyType("AkkaDeliveryExample1")
      auditOriginHostname("localhost")
      auditCriticality(Criticality.major)
      auditTS(new JDate())
      auditCreateTS(new JDate())
      auditUpdateTS(new JDate())
      auditMsg("message from akka delivery example #1")
      auditLabel("usersLabel1")
      auditData("test" -> Map("a" -> 1))
      auditData("java_version" -> System.getProperty("java.version"))
    }
  }

  Thread.sleep(3000L)

  AkkaDelivery.shutdown
}
