
/**
 * Copyright 2012-2013 Clint Combs
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.newhart.japi;

import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;
import org.newhart.*;
import static org.newhart.japi.NewhartJava.*;
import static scala.collection.JavaConversions.*;
import scala.Tuple2;
import java.io.PrintWriter;
import java.io.StringWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * JEntryBuilder is a Java class for building org.newhart.Entry instances.
 */
public class JEntryBuilder {

    private static final Logger logger = LoggerFactory.getLogger(JEntryBuilder.class);

    private String origin = "";
    private String originKey = "";
    private String originKeyType = "";
    private String originHostname = null;
    private JCriticality criticality = JCriticality.major;
    private Date auditTS = null;
    private Date createTS = null;
    private Date updateTS = null;
    private String msg = "";
    private Map<String, Object> data = new HashMap<String, Object>();
    private List<org.newhart.Error> errors = new LinkedList<org.newhart.Error>();
    private List<Warning> warnings = new LinkedList<Warning>();
    private Set<String> labels = new HashSet<String>();

    /**
     * Construct a new JEntryBuilder.
     *
     * @see org.newhart.japi.NewhartJava#entry()
     */
    public JEntryBuilder() {
    }

    public JEntryBuilder origin(final String s) {
        this.origin = s;
        return this;
    }

    public JEntryBuilder originKey(final String s) {
        this.originKey = s;
        return this;
    }

    public JEntryBuilder originKeyType(final String s) {
        this.originKeyType = s;
        return this;
    }

    public JEntryBuilder originKeyType(final Class c) {
        this.originKeyType = c.getName();
        return this;
    }

    /**
     * Set the originHostname.  If this is not called or is set to null then
     * org.newhart.package$.MODULE$.defaultOriginHostname() is used.  That method
     * returns the local hostname.
     */
    public JEntryBuilder originHostname(final String s) {
        this.originHostname = s;
        return this;
    }

    public JEntryBuilder criticality(final JCriticality c) {
        this.criticality = c;
        return this;
    }

    public JEntryBuilder auditTS(final Date d) {
        this.auditTS = d;
        return this;
    }

    public JEntryBuilder createTS(final Date d) {
        this.createTS = d;
        return this;
    }

    public JEntryBuilder updateTS(final Date d) {
        this.updateTS = d;
        return this;
    }

    public JEntryBuilder msg(final String s) {
        this.msg = s;
        return this;
    }

    public JEntryBuilder data(final String name, final Object value) {
        if (value != null) {
            data.put(name, value);
        } else {
            logger.error("data value must not be null for name: " + name);
        }
        return this;
    }

    public JEntryBuilder error(final org.newhart.Error e) {
        errors.add(e);
        return this;
    }

    public JEntryBuilder error(final String msg) {
        errors.add(buildError(msg));
        return this;
    }

    public JEntryBuilder error(final String msg, final String code) {
        errors.add(buildError(msg, code));
        return this;
    }

    public JEntryBuilder error(final String msg, final String code, final String stackTrace) {
        errors.add(buildError(msg, code, stackTrace));
        return this;
    }

    public JEntryBuilder error(final Throwable t) {
        errors.add(buildError(t.getMessage(), null, buildStackTraceString(t)));
        return this;
    }

    private String buildStackTraceString(final Throwable t) {
        return org.newhart.package$.MODULE$.buildStackTraceString(t);
    }

    public JEntryBuilder warning(final org.newhart.Warning w) {
        warnings.add(w);
        return this;
    }

    public JEntryBuilder warning(final String msg) {
        warnings.add(buildWarning(msg));
        return this;
    }

    public JEntryBuilder warning(final String msg, final String code) {
        warnings.add(buildWarning(msg, code));
        return this;
    }

    public JEntryBuilder warning(final String msg, final String code, final String stackTrace) {
        warnings.add(buildWarning(msg, code, stackTrace));
        return this;
    }

    public JEntryBuilder warning(final Throwable t) {
        warnings.add(buildWarning(t.getMessage(), null, buildStackTraceString(t)));
        return this;
    }

    public JEntryBuilder label(final String s) {
        labels.add(s);
        return this;
    }

    public Entry build() {
        final scala.collection.Seq<org.newhart.Error> scalaErrors = new JListWrapper<org.newhart.Error>(errors);
        final scala.collection.Seq<Warning> scalaWarnings = new JListWrapper<Warning>(warnings);
        final scala.collection.immutable.Set<String> scalaLabels = new JSetWrapper<String>(labels).toSet();

        scala.collection.immutable.HashMap<String, Object> scalaData =
            new scala.collection.immutable.HashMap<String, Object>();
        for(Map.Entry<String, Object> e:data.entrySet()) {
            final Tuple2<String, Object> keyValue =
                new Tuple2<String, Object>(e.getKey(), e.getValue());
            scalaData = scalaData.$plus(keyValue);
        }

        return new Entry(
            origin, originKey, originKeyType,
            (originHostname != null)?originHostname:org.newhart.package$.MODULE$.defaultOriginHostname(),
            Criticality.withName(criticality.name()),
            auditTS, createTS, updateTS,
            msg, scalaData, scalaErrors, scalaWarnings, scalaLabels);
    }
}
