
/**
 * Copyright 2012-2013 Clint Combs
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.newhart.mongodb

import com.mongodb.{Mongo,DBObject}
import com.mongodb.casbah.Imports._
import com.weiglewilczek.slf4s.Logging
import org.newhart.{Delivery,Entry,Error,Management,Retrieval,Warning}

/**
 * Companion object to the low-level API for working with audit stores.
 */
object AuditStore {
  def apply(mongo:Mongo):AuditStore = {
    implicit val con:MongoConnection = new MongoConnection(mongo)
    new AuditStore()
  }

  def instance(mongo:Mongo):AuditStore = {
    apply(mongo)
  }
}

trait MongoDelivery extends Delivery with NewhartMongo with Logging {

  @deprecated(message = "Please use send instead of writeEntry.", since = "0.4.5")
  def writeEntry(collectionName: String, e: Entry) = {
    send(collectionName, e)
  }

  def send(collectionName: String, e: Entry) = {
    try {
      con()(dbName)(collectionName) += MongoEntryBuilder.buildMongoEntryObject(e)
    } catch {
      case ex:Exception => logger.error("error writing entry: " + e, ex)
    }
  }
}

trait MongoRetrieval extends Retrieval with NewhartMongo with Logging {
  def search(name:String, q:DBObject, s:DBObject = MongoDBObject(), n:Option[Int] = Some(100)):Seq[Entry] = {
    // think about the return type and if the results should be streamed back or retrieved in pages

    val iter = n match {
      case Some(count) =>
        for(i <- con()(dbName)(name).find(q).sort(s).limit(count)) yield { parseEntry(i) }
      case None =>
        for(i <- con()(dbName)(name).find(q).sort(s)) yield { parseEntry(i) }
    }
    iter.toSeq
  }

  def searchIdEntryPairs(
    name:String, q:DBObject, s:DBObject = MongoDBObject(),
    n:Option[Int] = Some(100)): Seq[(ObjectId, Entry)] = {
    // think about the return type and if the results should be streamed back or retrieved in pages

    val iter = n match {
      case Some(count) =>
        for(i <- con()(dbName)(name).find(q).sort(s).limit(count)) yield {
          (i.getAsOrElse[ObjectId]("_id", new ObjectId()), parseEntry(i))
        }
      case None =>
        for(i <- con()(dbName)(name).find(q).sort(s)) yield {
          (i.getAsOrElse[ObjectId]("_id", new ObjectId()), parseEntry(i))
        }
    }
    iter.toSeq
  }

  private def parseEntry(dbObj:DBObject) = MongoEntryParser.parseMongoEntryObject(dbObj)

  /**
   * Retrieve the last 'n' entries in collection 'name'.
   *
   * TODO: consider changing this to return Iterator[Entry]...
   * it might be more efficient for large results
   */
  def last(name:String, n:Int):Seq[Entry] = {
    val iter = for(i <- con()(dbName)(name).find().sort(MongoDBObject("$natural" -> -1)).limit(n)) yield {
      MongoEntryParser.parseMongoEntryObject(i)
    }
    iter.toSeq
  }
}

trait MongoManagement extends Management with NewhartMongo with Logging {
  def dropAudit(name:String) = {
    con()(dbName)(name).dropCollection
  }

  def createAudit(name:String, size:Int) = {
    val db = con()(dbName)
    if (db.collectionNames contains name)
      db(name)
    else {
      val optionsBuilder = MongoDBObject.newBuilder
      optionsBuilder += "capped" -> true
      optionsBuilder += "size" -> size

      // index the collection by originKey, origin, and auditTS
      try {
        val col = db.createCollection(name, optionsBuilder.result)
        col.ensureIndex(MongoDBObject("originKey" -> 1, "origin" -> 1, "auditTS" -> 1))
        logger.info("created collection in newhart database '" + name + "' of size " + size)
      } catch {
        case ex:Exception => logger.error("error creating collection '" + name + "' of size " + size)
      }
    }
  }
}

/**
 * AuditStore is the original base API for directly manipulating audit collections and
 * entries in MongoDB.  It is now being retrofitted to extend the new Delivery, Management,
 * and Retrieval traits.  It may be deprecated in the near future.
 */
class AuditStore(implicit mongoConnection: MongoConnection)
extends MongoDelivery with MongoManagement with MongoRetrieval {

  def con(): MongoConnection = mongoConnection
}
